# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import ServiceHub, ServiceDisabledError, ServiceUnavailableError, InstanceDefault

import ldap3
import ldap3.core.exceptions as ldapexc
import ldap3.utils.config as ldap3_conf
import re
import threading
import yaml

from ldap3.utils.conv import escape_filter_chars
import logging
logger = logging.getLogger("ldap")

# Reduce block time when LDAP server is not reachable
ldap3_conf.set_config_parameter("RESTARTABLE_SLEEPTIME", 1)
ldap3_conf.set_config_parameter("RESTARTABLE_TRIES", 2)


def handleLdapError(service, error):
    if isinstance(error,
                  (ldapexc.LDAPSocketOpenError, ldapexc.LDAPSocketSendError, ldapexc.LDAPSessionTerminatedByServerError,
                   ldapexc.LDAPMaximumRetriesError)):
        return ServiceHub.SUSPENDED


def argname(orgID=None):
    if orgID is not None:
        from orm.domains import Orgs
        org = Orgs.query.filter(Orgs.ID == orgID).with_entities(Orgs.name).first()
        return org.name if org else None


def orgid(orgspec):
    if orgspec is None:
        return 0
    try:
        return int(orgspec)
    except ValueError:
        from orm.domains import Orgs
        org = Orgs.query.filter(Orgs.name == orgspec).with_entities(Orgs.ID).first()
        if org is None:
            raise
        return org.ID


class SearchResult:
    def __init__(self, ldap, resultType, data):
        self._ldap = ldap
        userconf, groupconf = ldap._config["users"], ldap._config["groups"]
        self.DN = data.get("dn")
        self.type = resultType
        self.error = None
        if "raw_attributes" in data and "attributes" in data:
            self.ID = data["raw_attributes"][ldap._config["objectID"]][0]
            self.data = data["attributes"]
            self.name = self._reduce(data["attributes"].get(userconf["displayName"], ""))
        else:
            self.ID = self.data = self.name = self.email = None
            self.error = "Not a valid object"
            return
        if resultType == "user":
            if userconf["username"] in data["attributes"] and data["attributes"][userconf["username"]]:
                self.email = self.username = self._reduce(data["attributes"][userconf["username"]])
            else:
                self.email = self.username = None
                self.error = "Missing username"
        elif resultType == "contact":
            self.username = None
            if userconf["contactname"] in data["attributes"] and data["attributes"][userconf["contactname"]]:
                self.email = self._reduce(data["attributes"][userconf["contactname"]])
            else:
                self.email = None
                self.error = "Missing e-mail address"
        elif resultType == "group":
            self.email = self._reduce(data["attributes"].get(groupconf["groupaddr"], ""))
            self.name = self._reduce(data["attributes"].get(groupconf["groupname"], ""))
            if not self.email:
                self.error = "Missing e-mail address"
        else:
            self.error = "Unknown type"

    def __repr__(self):
        return "<{} {}>".format(self.type, self.email)

    def _chkConvert(self, requested):
        if self.error:
            raise ValueError("Cannot create {} data: {}".format(requested, self.error))
        if self.type != requested:
            raise TypeError("Cannot create {} data from {} object".format(requested, self.type))

    @staticmethod
    def _reduce(value, tail=False):
        """Reduce potentially multi-valued attribute to scalar.

        Parameters
        ----------
        value : Any
            Value to reduce
        tail : bool, optional
            Return additional elements in second return value. The default is False.

        Returns
        -------
        Any
            Scalar value (tail=False) or two-tuple containing the scalar and additional elements (tail=True).
        """
        if isinstance(value, (list, tuple)):
            res = (value[0], value[1:]) if value else (None, [])
        else:
            res = (value, [])
        return res if tail else res[0]

    def userdata(self, props=None):
        if self.type == "contact":
            return self.contactdata(props)
        elif self.type == "group":
            return self.groupdata(props)
        self._chkConvert("user")
        ldap = self._ldap
        ldapuser = self.data
        username, aliases = self._reduce(ldapuser[ldap._config["users"]["username"]], tail=True)
        userdata = dict(username=username.lower(), aliases=aliases)
        userdata["properties"] = props or ldap._defaultProps.copy()
        userdata["properties"].update({prop: " ".join(str(a) for a in ldapuser[attr]) if isinstance(ldapuser[attr], list)
                                       else ldapuser[attr] for attr, prop in ldap._userAttributes.items() if attr in ldapuser})
        if ldap._config["users"].get("aliases"):
            aliasattr = ldap._config["users"]["aliases"]
            if ldapuser.get(aliasattr) is not None:
                from tools import formats
                aliases = ldapuser[aliasattr]
                aliases = aliases if isinstance(aliases, list) else [aliases]
                aliases = [alias[5:] if alias.lower().startswith("smtp:") else alias for alias in aliases]
                userdata["aliases"] += [alias for alias in aliases if formats.email.match(alias)]
        return userdata

    def contactdata(self, props=None):
        self._chkConvert("contact")
        ldap = self._ldap
        ldapuser = self.data
        userdata = {}
        userdata["properties"] = props or ldap._defaultProps.copy()
        userdata["properties"].update({prop: " ".join(str(a) for a in ldapuser[attr]) if isinstance(ldapuser[attr], list)
                                       else ldapuser[attr] for attr, prop in ldap._userAttributes.items() if attr in ldapuser})
        return userdata

    def groupdata(self, props=None):
        self._chkConvert("group")
        return dict(listname=self.email, listType=0, displayname=self.name)


@ServiceHub.register("ldap", handleLdapError, maxreloads=3, argspec=((), (orgid,)), argname=argname)
class LdapService:
    __initialized = False
    _templates = {}
    _unescapeRe = re.compile(rb"\\(?P<value>[a-fA-F0-9]{2})")

    _configMap = {"baseDn": "ldap_search_base",
                  "objectID": "ldap_object_id",
                  "users": "ldap_mail_attr",
                  "connection": "ldap_host",
                  "username": "ldap_mail_attr",
                  "searchAttributes": "ldap_user_search_attrs",
                  "displayName": "ldap_user_displayname",
                  "groupaddr": "ldap_group_addr",
                  "groupfilter": "ldap_group_filter",
                  "groupname": "ldap_group_name"}

    @classmethod
    def init(cls):
        if cls.__initialized:
            return
        ldap3.set_config_parameter("POOLING_LOOP_TIMEOUT", 1)
        try:
            with open("res/ldapTemplates.yaml", encoding="utf-8") as file:
                cls._templates = yaml.load(file, Loader=yaml.SafeLoader)
        except Exception:
            pass
        cls.__initialized = True

    def __init__(self, orgID=None):
        from tools import mconf
        self.init()
        if orgID is None:
            self._config = mconf.LDAP
        else:
            self._config = self._loadOrgConfig(orgID)
        self._userAttributes = self._checkConfig(self._config)
        self.lock = threading.Lock()
        if self._config.get("disabled"):
            raise ServiceDisabledError("Service disabled by configuration")
        try:
            self.conn = self.testConnection(self._config)
        except ldap3.core.exceptions.LDAPInvalidDnError:
            raise ServiceUnavailableError("Invalid base DN")
        except Exception as err:
            msg = str(err.args[0]) if len(err.args) else type(err).__name__
            raise ServiceUnavailableError("Failed to connect to server: "+msg, *err.args[1:])
        if "defaultQuota" in self._config["users"]:
            self._defaultProps = {prop: self._config["users"]["defaultQuota"] for prop in
                                  ("storagequotalimit", "prohibitsendquota", "prohibitreceivequota")}
        else:
            self._defaultProps = {}

    def _attrSet(self, name, mode="user"):
        if isinstance(name, (list, tuple)):
            return name
        common = (self._config["objectID"],)
        if name == "idonly":
            return common
        if name == "all":
            return common+("*",)
        if mode == "group":
            groupconf = self._config["groups"]
            return common+(groupconf["groupaddr"], groupconf["groupname"])
        userconf = self._config["users"]
        common += (userconf["displayName"],)
        if mode == "user":
            return common+(userconf["username"],)
        elif mode == "contact":
            return (common+(userconf["contactname"],)) if self._config["enableContacts"] else common

    @classmethod
    def _checkConfig(cls, config):
        for required in ("baseDn", "objectID", "users", "connection"):
            if required not in config or config[required] is None or len(config[required]) == 0:
                raise KeyError("Missing required config value '{}'".format(cls._configMap.get(required, required)))
        if "server" not in config["connection"] or config["connection"]["server"] is None or\
           len(config["connection"]["server"]) == 0:
            raise KeyError("Missing required config value 'ldap_host'")
        for required in ("username", "searchAttributes", "displayName"):
            if required not in config["users"] or config["users"][required] is None or len(config["users"][required]) == 0:
                raise KeyError("Missing required config value '{}'".format(cls._configMap.get(required, "user."+required)))
        _templatesEnabled = config["users"].get("templates", [])
        userAttributes = {}
        if "filter" in config["users"]:
            f = config["users"]["filter"]
            if f is not None and len(f) != 0 and f[0] != "(" and f[-1] != ")":
                config["users"]["filter"] = "("+f+")"
        for _template in _templatesEnabled:
            if _template not in cls._templates:
                raise ValueError("Unknown template '{}'".format(_template))
            userAttributes.update(cls._templates.get(_template, {}))
        userAttributes.update(config["users"].get("attributes", {}))
        userAttributes[config["users"]["displayName"]] = "displayname"
        for key, value in userAttributes.items():
            if value == "smtpaddress":
                config["users"]["contactname"] = key
                break
        config["enableContacts"] = bool(config["users"].get("contactFilter") and config["users"].get("contactname"))
        if config["groups"]:
            for required in ("groupaddr", "groupfilter", "groupname"):
                if required not in config["groups"] or not config["groups"][required]:
                    raise KeyError("Missing required config value '{}'"
                                   .format(cls._configMap.get(required, "groups."+required)))
        return userAttributes

    def _matchFilters(self, ID):
        """Generate match filters string.

        Includes a filter for each entry in ldap.users.filters and adds ID filter.

        Parameters
        ----------
        ID : str or bytes
            Object ID of the LDAP person

        Returns
        -------
        str
            A string containing LDAP match filter expression.
        """
        return "({}={})".format(self._config["objectID"], self.escape_filter_chars(ID))

    def _matchFiltersMulti(self, IDs):
        """Generate match filters string for multiple IDs.

        Includes a filter for each entry in ldap.users.filters and adds ID filters.

        Parameters
        ----------
        IDs : list of bytes or str
            List of IDs to match

        Returns
        -------
        str
            A string containing LDAP match filter expression.
        """
        return "(|{})".format("".join("({}={})".format(self._config["objectID"], self.escape_filter_chars(ID)) for ID in IDs))

    @property
    def _sbase(self):
        return self._searchBase(self._config)

    def _search(self, baseFilter, *args, attributes=None, domains=None, filterIncomplete=True, limit=None, userconf=None,
                types=None, customFilter="", **kwargs):
        """Perform async search query.

        Parameters
        ----------
        *args : Any
            Arguments forwarded to conn.search
        **kwargs : Any
            Keyword arguments forwarded to conn.search

        Returns
        -------
        list
            Search result list
        """
        def filtered(results):
            return list(filter(lambda r: r.error is None, results)) if filterIncomplete else list(results)

        def searchPaged(typeFilter, type, *args, **kwargs):
            filterExpr = "(&{}{}{})".format(baseFilter, typeFilter, customFilter)
            if not self.conn.search(self._sbase, filterExpr, *args, **kwargs):
                return []
            results = filtered(SearchResult(self, type, result) for result in self.conn.response)
            cookie = self.conn.result.get("controls", {}).get("1.2.840.113556.1.4.319", {}).get("value", {}).get("cookie")
            while cookie and (not limit or len(results) < limit) and \
                  self.conn.search(self._sbase, filterExpr, *args, **kwargs, paged_cookie=cookie):
                results += filtered(SearchResult(self, type, result) for result in self.conn.response)
                cookie = self.conn.result.get("controls", {}).get("1.2.840.113556.1.4.319", {}).get("value", {}).get("cookie")
            return results[:limit] if limit and len(results) > limit else results

        if limit:
            kwargs["paged_size"] = min(limit, kwargs.get("paged_size") or limit)
        types = types or ("user", "contact", "group")
        userconf = userconf or self._config["users"]
        username = userconf["username"]
        domainexpr = "(|{})".format("".join("({}=*@{})".format(username, d) for d in domains)) if domains is not None else ""
        filterexpr = "".join("("+f+")" for f in userconf.get("filters", ()))
        userFilter = "(&{}{}{})".format(filterexpr, userconf.get("filter", ""), domainexpr)
        with self.lock:
            results = []
            if "user" in types:
                results += searchPaged(userFilter, "user", *args, attributes=self._attrSet(attributes, "user"), **kwargs)
                limit = None if limit is None else limit-len(results)
            if self._config["enableContacts"] and "contact" in types and (limit is None or limit > 0):
                contacts = searchPaged(self._config["users"]["contactFilter"], "contact", *args,
                                       attributes=self._attrSet(attributes, "contact"), **kwargs)
                results += contacts
                limit = None if limit is None else limit-len(contacts)
            if self._config["groups"] and "group" in types and (limit is None or limit > 0):
                results += searchPaged(self._config["groups"]["groupfilter"], "group", *args,
                                       attributes=self._attrSet(attributes, "group"), **kwargs)
        return results

    @classmethod
    def _searchBase(cls, conf):
        """Generate directory name to search.

        If configured, adds the ldap.users.subtree path to ldap.baseDn. Otherwise only ldap.baseDn is returned.

        Returns
        -------
        str
            LDAP directory to search for users.
        """
        if "users" in conf and "subtree" in conf["users"]:
            return conf["users"]["subtree"]+","+conf["baseDn"]
        return conf["baseDn"]

    @classmethod
    def _searchFilters(cls, query, userconf, domains=None):
        """Generate search filters string.

        Includes a filter for each entry in ldap.users.filters and adds substring filters for all attributes in
        ldap.users.searchAttributes.
        Optionally, an additional list of permitted domains can be used to further restrict matches.

        Parameters
        ----------
        query : str
            Username to search for.

        Returns
        -------
        str
            A string including all search filters.
        """
        if query:
            query = cls.escape_filter_chars(query)
            return "(|{})".format("".join(("("+sattr+"=*"+query+"*)" for sattr in userconf["searchAttributes"])))
        else:
            return ""

    @classmethod
    def _loadOrgConfig(cls, orgID):
        from orm.domains import OrgParam
        config = OrgParam.loadLdap(orgID)
        if config is None:
            raise InstanceDefault()
        return config

    def authUser(self, ID, password):
        """Attempt ldap bind for user with given ID and password

        Parameters
        ----------
        ID : str or bytes
            ID of the LDAP object representing the user
        password : str
            User password.

        Returns
        -------
        str
            Error message if authentication failed or None if successful
        """
        response = self._search(self._matchFilters(ID), attributes="idonly", filterIncomplete=False)
        if len(response) == 0:
            return "Invalid Username or password"
        if len(response) > 1:
            return "Multiple entries found - please contact your administrator"
        userDN = response[0].DN
        try:
            ldap3.Connection(self._config["connection"].get("server"), user=userDN, password=password, auto_bind=True)
        except ldapexc.LDAPBindError:
            return "Invalid username or Password"

    def downsyncUser(self, ID, props=None):
        """Create dictionary representation of the user from LDAP data.

        The dictionary can be used to create or update a orm.users.Users object.

        Parameters
        ----------
        ID : str
            LDAP ID of the user object
        props : dict, optional
            UserProperties as dictionary. The default is a dictionary containing storagequotalimit property.

        Raises
        ------
        RuntimeError
            LDAP query failed

        Returns
        -------
        userdata : dict
            Dictionary representation of the LDAP user
        """
        try:
            response = self._search(self._matchFilters(ID), attributes="all")
        except Exception:
            return None
        if len(response) == 0:
            return None
        if len(response) > 1:
            raise RuntimeError("Multiple entries found - aborting")
        return response[0].userdata()

    def dumpUser(self, ID):
        """Download complete user description.

        Parameters
        ----------
        ID : str ot bytes
            LDAP object ID of the user

        Returns
        -------
        ldap3.abstract.entry.Entry
            LDAP object or None if not found or ambiguous
        """
        res = self._search(self._matchFilters(ID), attributes="all")
        if len(res) != 1:
            return None
        res = res[0]
        return yaml.dump({"DN": res.DN})+yaml.dump({"attributes": dict(res.data)})

    @staticmethod
    def escape_filter_chars(text, encoding=None):
        return escape_filter_chars(text, encoding)

    def getAll(self, IDs):
        """Get user information for each ID.

        Queries the same information as getUserInfo.

        Parameters
        ----------
        IDs : list of bytes or str
            IDs to search

        Returns
        -------
        list
            List of GenericObjects with information about found users
        """
        return self._search(self._matchFiltersMulti(IDs))

    def getUserInfo(self, ID):
        """Get e-mail address of an ldap user.

        Parameters
        ----------
        ID : str or bytes
            Object ID of the LDAP user
        Returns
        -------
        GenericObject
            Object containing LDAP ID, username and display name of the user
        """
        try:
            response = self._search(self._matchFilters(ID))
        except ldapexc.LDAPInvalidValueError:
            return None
        if len(response) != 1:
            return None
        return response[0]

    def groupMemberFilter(self, groupDN):
        """Generate filter expression for group members.

        Parameters
        ----------
        groupDN : str
            DN of the group

        Returns
        -------
        str
            Filter expression
        """
        return "({}={})".format(self._config["groups"].get("groupMemberAttr", "memberOf"), groupDN)

    def searchUsers(self, query=None, domains=None, limit=None, pageSize=1000, filterIncomplete=True, types=None,
                    customFilter="", attributes=None):
        """Search for ldap users matching the query.

        Parameters
        ----------
        query : str
            String to match
        domains : list of str, optional
            Optional domain filter. The default is None.
        limit : int, optional
            Maximum number of results to return or None for no limit. Default is None.
        pageSize : int, optional
            Perform a paged search with given page size. Default is 1000.
        types: iterable, optional
            Only search for specified types. Default is ("users", "contacts", "groups")
        customFilter: str, optional
            Custom filter expression to add to the search. Default is ""

        Returns
        -------
        list
            List of user objects containing ID, e-mail and name
        """
        try:
            exact = self.getUserInfo(self.unescapeFilterChars(query))
            exact = [] if exact is None else [exact]
        except Exception:
            exact = []
        response = self._search(self._searchFilters(query, self._config["users"], domains),
                                domains=domains,
                                paged_size=pageSize,
                                limit=limit,
                                filterIncomplete=filterIncomplete,
                                types=types,
                                customFilter=customFilter)
        return exact+response

    @classmethod
    def testConfig(cls, config):
        cls.init()
        try:
            cls._checkConfig(config)
            cls.testConnection(config)
        except KeyError as err:
            return "Incomplete LDAP configuration: "+err.args[0]
        except ValueError as err:
            return "Invalid LDAP configuration: "+err.args[0]
        except ldapexc.LDAPAttributeError as err:
            return "Bad attribute configuration: "+err.args[0]
        except Exception as err:
            return "Could not connect to LDAP server: "+" - ".join(str(v) for v in err.args)

    @classmethod
    def testConnection(cls, config, active=True):
        servers = [s[:-1] if s.endswith("/") else s for s in config["connection"]["server"].split()]
        pool = servers[0] if len(servers) == 1 else ldap3.ServerPool(servers, "FIRST", active=1)
        user = config["connection"].get("bindUser")
        password = config["connection"].get("bindPass")
        starttls = config["connection"].get("starttls")
        conn = ldap3.Connection(pool, user=user, password=password, client_strategy=ldap3.RESTARTABLE)
        if starttls and not conn.start_tls():
            logger.warning("Failed to initiate StartTLS connection")
        if not conn.bind():
            raise ldapexc.LDAPBindError("LDAP bind failed ({}): {}".format(conn.result["description"], conn.result["message"]))
        if active:
            userconf = config["users"]
            filterexpr = "".join("("+f+")" for f in userconf.get("filters", ()))
            userFilter = "(&{}{})".format(filterexpr, userconf.get("filter", ""))
            attrs = (config["objectID"], userconf["displayName"], userconf["username"])
            conn.search(cls._searchBase(config), userFilter, attributes=attrs, paged_size=0)
            if config.get("enableContacts"):
                contactFilter = "(&{}{})".format(filterexpr, userconf.get("contactFilter", ""))
                attrs = (config["objectID"], userconf["displayName"], userconf["contactname"])
                conn.search(cls._searchBase(config), contactFilter, attributes=attrs, paged_size=0)
        return conn

    @classmethod
    def unescapeFilterChars(cls, text):
        """Reverse escape_filter_chars function.

        In contrast to ldap3.utils.conv.unescape_filter_chars, this function also processes arbitrary byte escapes.

        Parameters
        ----------
        text : str
            String generated by ldap3.utils.conv.escape_filter_chars


        Returns
        -------
        bytes
            bytes object containing unescaped data
        """
        raw = bytes(text, "utf-8")
        last = 0
        unescaped = bytes()
        for match in cls._unescapeRe.finditer(raw):
            unescaped += raw[last:match.start()]+bytes((int(match.group("value"), 16),))
            last = match.end()
        return unescaped if last != 0 else raw



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from flask import request, jsonify
from sqlalchemy.exc import IntegrityError

from .. import defaultListHandler, defaultObjectHandler, defaultPatch

import api
from api.core import API, secure
from api.security import checkPermissions

from services import ServiceHub, Service
from services.ldap import LdapService

from tools.permissions import SystemAdminPermission, SystemAdminROPermission
from tools.permissions import DomainAdminROPermission, OrgAdminPermission, DomainPurgePermission

from orm import DB


@API.route(api.BaseRoute+"/system/orgs", methods=["GET", "POST"])
@secure(requireDB=True)
def orgListEndpoint():
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    from orm.domains import Orgs
    return defaultListHandler(Orgs)


@API.route(api.BaseRoute+"/system/orgs/<int:ID>", methods=["GET", "PATCH"])
@secure(requireDB=True)
def orgObjectEndpoint(ID):
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    from orm.domains import Orgs
    return defaultObjectHandler(Orgs, ID, "Organization")


@API.route(api.BaseRoute+"/system/orgs/<int:ID>", methods=["DELETE"])
@secure(requireDB=True)
def orgDeleteEndpoint(ID):
    checkPermissions(SystemAdminPermission())
    from orm.domains import Domains, Orgs
    Domains.query.filter(Domains.orgID == ID).update({Domains.orgID: 0}, synchronize_session=False)
    return defaultObjectHandler(Orgs, ID, "Organization")


###################################################################################################

@API.route(api.BaseRoute+"/system/orgs/<int:ID>/ldap", methods=["GET"])
@secure(requireDB=109)
def getOrgLdapConfig(ID):
    checkPermissions(OrgAdminPermission(ID))
    from orm.domains import OrgParam
    config = OrgParam.loadLdap(ID) or {}
    return jsonify(ldapAvailable=Service.available("ldap", ID) if config else False, data=config)


@API.route(api.BaseRoute+"/system/orgs/<int:ID>/ldap", methods=["PUT"])
@secure(requireDB=109)
def updateOrgLdapConfig(ID):
    checkPermissions(OrgAdminPermission(ID))
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing configuration"), 400
    error = LdapService.testConfig(data)
    forced = False
    if error:
        if request.args.get("force") == "true":
            forced = True
        else:
            return jsonify(message=error), 400
    from orm.domains import OrgParam
    OrgParam.saveLdap(ID, data)
    ServiceHub.load("ldap", ID, force_reload=True)
    return jsonify(message="Force updated LDAP configuration" if forced else "LDAP configuration updated")


@API.route(api.BaseRoute+"/system/orgs/<int:ID>/ldap", methods=["DELETE"])
@secure(requireDB=109)
def deleteOrgLdapConfig(ID):
    checkPermissions(OrgAdminPermission(ID))
    from orm.domains import OrgParam
    OrgParam.wipeLdap(ID)
    ServiceHub.load("ldap", ID, force_reload=True)
    return jsonify(message="Success.")

###################################################################################################

@API.route(api.BaseRoute+"/system/domains", methods=["GET"])
@secure(requireDB=True)
def domainListEndpoint():
    checkPermissions(SystemAdminROPermission())
    from orm.domains import Domains
    return defaultListHandler(Domains)


@API.route(api.BaseRoute+"/system/domains", methods=["POST"])
@secure(requireDB=True)
def domainCreate():
    checkPermissions(SystemAdminPermission())
    from orm.domains import Domains
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing data"), 400
    if SystemAdminPermission not in request.auth["user"].permissions():
        data.pop("homeserver", None)
    result, code = Domains.create(data, request.args.get("createRole", "false") == "true")
    if code != 201:
        return jsonify(message=result), code
    return jsonify(result.fulldesc()), 201


@API.route(api.BaseRoute+"/system/domains/<int:domainID>", methods=["GET"])
@secure(requireDB=True)
def getDomain(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.domains import Domains
    return defaultObjectHandler(Domains, domainID, "Domain")


@API.route(api.BaseRoute+"/system/domains/<int:domainID>", methods=["PATCH"])
@secure(requireDB=True)
def updateDomain(domainID):
    checkPermissions(OrgAdminPermission("*"))
    from orm.domains import Domains
    from orm.users import Users
    domain: Domains = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    checkPermissions(OrgAdminPermission(domain.orgID))
    oldStatus = domain.domainStatus
    patched = defaultPatch(Domains, domainID, "Domain", obj=domain, result="precommit")
    if isinstance(patched, tuple):  # Return value is not the domain, but an error response
        return patched
    if oldStatus != domain.domainStatus:
        Users.query.filter(Users.domainID == domainID)\
                   .update({Users.addressStatus: Users.addressStatus.op("&")(0xF)+(domain.domainStatus << 4)},
                           synchronize_session=False)
    try:
        DB.session.commit()
    except IntegrityError as err:
        DB.session.rollback()
        return jsonify(message="Domain update failed", error=err.orig.args[1])
    return jsonify(domain.fulldesc())


@API.route(api.BaseRoute+"/system/domains/<int:domainID>", methods=["DELETE"])
@secure(requireDB=True)
def deleteDomain(domainID):
    checkPermissions(OrgAdminPermission("*"))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    checkPermissions(OrgAdminPermission(domain.orgID))
    if request.args.get("purge") == "true":
        checkPermissions(DomainPurgePermission())
        domain.purge(request.args.get("deleteFiles") == "true")
        msg = "removed."
    else:
        domain.delete()
        msg = "marked as deleted."
    DB.session.commit()
    return jsonify(message="Domain "+msg)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from api.core import API
from flask import jsonify, request
from services import ServiceUnavailableError
from sqlalchemy.exc import DatabaseError

import traceback


class InsufficientPermissions(RuntimeError):
    def __init__(self, msg=None):
        RuntimeError(self, msg)


@API.errorhandler(DatabaseError)
def database_error(error):
    API.logger.error("Database query failed: {}".format(error))
    return jsonify(message="Database error."), 503


@API.errorhandler(InsufficientPermissions)
def insufficient_permissions(error):
    msg = error.args[0] if len(error.args) else "Insufficient permissions for this operation"
    return jsonify(message="Access denied: "+msg), 403


@API.errorhandler(ServiceUnavailableError)
def service_unavailable(error):
    return jsonify(message=error.args[0]), 503


@API.errorhandler(Exception)
def internal_error(error):
    """Return JSON object with 405 message."""
    API.logger.error(traceback.format_exc())
    return jsonify(message="The server encountered an error while processing the request."), 500


@API.errorhandler(404)
def not_found(error):
    """Return JSON object with 404 message."""
    return jsonify(message="Resource not found"), 404


@API.errorhandler(405)
def method_not_allowed(error):
    """Return JSON object with 405 message."""
    return jsonify(message="Method '{}' not allowed on this endpoint".format(request.method)), 405


@API.after_request
def logError(response):
    if response.status_code // 100 in (4, 5):
        API.logger.warning("{} {} from {} -> {} {}".format(request.method, request.full_path, request.remote_addr,
                                                           response.status_code, repr(str(response.data, "utf-8"))
                                                           if response.is_json else "<data>"))
    return response



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020-2021 grommunio GmbH

import api

from api.core import API, secure
from api.security import checkPermissions
from base64 import b64decode
from datetime import datetime
from flask import request, jsonify
from sqlalchemy.exc import IntegrityError

from .. import defaultObjectHandler, userQuery

from services import Service

from tools import formats
from tools.config import Config
from tools.constants import PropTags, PropTypes, ExchangeErrors, PrivateFIDs, Permissions
from tools.misc import loadPSO, GenericObject
from tools.permissions import SystemAdminPermission, DomainAdminPermission, DomainAdminROPermission
from tools.rop import nxTime, makeEidEx
from tools.storage import setDirectoryOwner, setDirectoryPermission

import configparser
import json
import os
import shutil
import time

from orm import DB


@API.route(api.BaseRoute+"/domains/<int:domainID>/users", methods=["GET"])
@secure(requireDB=True)
def getUsers(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    return userQuery(domainID)


@API.route(api.BaseRoute+"/domains/<int:domainID>/users", methods=["POST"])
@secure(requireDB=True, authLevel="user")
def createUser(domainID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users
    data = request.get_json(silent=True) or {}
    data["domainID"] = domainID
    if SystemAdminPermission() not in request.auth["user"].permissions():
        data.pop("homeserver", None)
    isContact = data.get("status") == Users.CONTACT
    result, code = Users.mkContact(data) if isContact else Users.create(data)
    if code != 201:
        return jsonify(message=result), code
    return jsonify(result.fulldesc()), 201


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>", methods=["GET", "PATCH"])
@secure(requireDB=True, authLevel="user")
def userObjectEndpoint(domainID, userID):
    checkPermissions(DomainAdminROPermission(domainID) if request.method == "GET" else DomainAdminPermission(domainID))
    if request.method == "PATCH" and SystemAdminPermission() not in request.auth["user"].permissions():
        data = request.get_json(silent=True) or {}
        data.pop("homeserver", None)
    from orm.users import Users
    return defaultObjectHandler(Users, userID, "User", filters=(Users.domainID == domainID,))


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>", methods=["DELETE"])
@secure(requireDB=True)
def deleteUserEndpoint(domainID, userID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User #{} not found".format(userID)), 404
    return deleteUser(user, not request.args.get("deleteChatUser") == "false")


def deleteUser(user, deleteChatUser):
    if user.ID == 0:
        return jsonify(message="Cannot delete superuser"), 400
    userdata = GenericObject(maildir=user.maildir, homeserver=user.homeserver)
    user.delete(deleteChatUser)
    try:
        DB.session.commit()
    except Exception:
        return jsonify(message="Cannot delete user: Database commit failed."), 500
    if userdata.maildir:
        with Service("exmdb", errors=Service.SUPPRESS_INOP) as exmdb:
            client = exmdb.user(userdata)
            client.unloadStore()
    if request.args.get("deleteFiles") == "true":
        shutil.rmtree(userdata.maildir, ignore_errors=True)
    return jsonify(message="isded")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/password", methods=["PUT"])
@secure(requireDB=True, authLevel="user")
def setUserPassword(domainID, userID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users
    if userID == request.auth["user"].ID:
        return jsonify(message="Cannot reset own password, use '/passwd' endpoint instead"), 400
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if user.externID is not None:
        return jsonify(message="Cannot modify LDAP imported user"), 400
    data = request.get_json(silent=True)
    if data is None or "new" not in data:
        return jsonify(message="Incomplete data"), 400
    user.password = data["new"]
    DB.session.commit()
    return jsonify(message="Success")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/roles", methods=["PATCH"])
@secure(requireDB=True)
def updateUserRoles(domainID, userID):
    checkPermissions(SystemAdminPermission())
    from orm.roles import AdminUserRoleRelation, AdminRoles
    data = request.get_json(silent=True)
    if data is None or "roles" not in data:
        return jsonify(message="Missing roles array"), 400
    roles = {role.roleID for role in AdminUserRoleRelation.query.filter(AdminUserRoleRelation.userID == userID).all()}
    requested = set(data["roles"])
    remove = roles-requested
    add = requested-roles
    AdminUserRoleRelation.query.filter(AdminUserRoleRelation.userID == userID, AdminUserRoleRelation.roleID.in_(remove))\
                               .delete(synchronize_session=False)
    for ID in add:
        DB.session.add(AdminUserRoleRelation(userID, ID))
    try:
        DB.session.commit()
    except IntegrityError as err:
        return jsonify(message="Invalid data", error=err.orig.args[1]), 400
    roles = AdminRoles.query.join(AdminUserRoleRelation).filter(AdminUserRoleRelation.userID == userID).all()
    return jsonify(data=[role.ref() for role in roles])


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/storeProps", methods=["GET", "DELETE"])
@secure(requireDB=True)
def rdUserStoreProps(domainID, userID):
    checkPermissions(DomainAdminROPermission(domainID) if request.method == "GET" else DomainAdminPermission(domainID))
    from orm.users import DB, Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    props = [prop for prop in request.args.get("properties", "").split(",") if prop != ""]
    try:
        user.properties = {prop: None for prop in props}
    except ValueError as err:
        return jsonify(message=err.args[0])
    if len(props) == 0:
        return jsonify(data={}) if request.method == "GET" else jsonify(message="Nothing to delete")
    for i in range(len(props)):
        if not hasattr(PropTags, props[i].upper()) or not isinstance(getattr(PropTags, props[i].upper()), int):
            return jsonify(message="Unknown property '{}'".format(props[i])), 400
        props[i] = getattr(PropTags, props[i].upper())
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        if request.method == "DELETE":
            client.removeStoreProperties(props)
            DB.session.commit()
            return jsonify(message="Success.")
        propvals = client.getStoreProperties(0, props)
    respData = {}
    for propval in propvals:
        propname = PropTags.lookup(propval.tag).lower()
        if propval.tag & 0xFFFF == PropTypes.FILETIME:
            respData[propname] = datetime.fromtimestamp(nxTime(propval.val)).strftime("%Y-%m-%d %H:%M:%S")
        else:
            respData[propname] = propval.val
    return jsonify(data=respData)


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/storeProps", methods=["PATCH"])
@secure(requireDB=True)
def setUserStoreProps(domainID, userID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import DB, Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    data = request.get_json(silent=True)
    if data is None or len(data) == 0:
        return jsonify(message="Missing data"), 400
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    errors = {}
    propvals = []
    with Service("exmdb") as exmdb:
        for prop, val in data.items():
            try:
                tag = PropTags.deriveTag(prop)
            except ValueError:
                errors[prop] = "Unknown tag"
                continue
            try:
                val = PropTags.convertValue(tag, val)
            except ValueError:
                errors[prop] = "Invalid type"
                continue
            try:
                propvals.append(exmdb.TaggedPropval(tag, val))
            except TypeError:
                errors[prop] = "Unsupported type"
        client = exmdb.user(user)
        problems = client.setStoreProperties(0, propvals)
        for entry in problems:
            tag = PropTags.lookup(entry.proptag, hex(entry.proptag)).lower()
            err = ExchangeErrors.lookup(entry.err, hex(entry.err))
            errors[tag] = err
        user.properties = data
        DB.session.commit()
        if len(errors) != 0:
            API.logger.warn("Failed to set proptags: "+", ".join("{} ({})".format(tag, err) for tag, err in errors.items()))
        return jsonify(message="Great success!" if len(errors) == 0 else "Some tags could not be set", errors=errors)


def decodeSyncState(data, username):
    data = b64decode(data)
    if len(data) >= 2 and data[1] == ord(":"):
        API.logger.warning("Loading PHP serialize objects is deprecated")
        return loadPSO(data, decode_strings=True)["StateObject"][1]["devices"][username]["ASDevice"][1]
    elif len(data) >= 1 and data[0] == ord("{"):
        data = json.loads(data)["data"]
        if "devices" in data:
            data = data["devices"][username]["data"]
        return data
    return None


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sync", methods=["GET"])
@secure(requireDB=True)
def getUserSyncData(domainID, userID):
    checkPermissions(DomainAdminROPermission(domainID))
    props = ("deviceid", "devicetype", "useragent", "deviceuser", "firstsynctime", "lastupdatetime", "asversion")
    from orm.users import DB, Users, UserDevices
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    with Service("exmdb") as exmdb:
        devices = {}
        client = exmdb.user(user)
        data = client.getSyncData(Config["sync"].get("syncStateFolder", "GS-SyncState"))
    for device, state in data.items():
        try:
            stateobj = decodeSyncState(state, user.username)
            if stateobj is None:
                continue
            syncstate = {prop: stateobj[prop] for prop in props}
            syncstate["foldersSyncable"] = len(stateobj["contentdata"])
            syncstate["foldersSynced"] = len([folder for folder in stateobj["contentdata"].values() if 1 in folder])
            syncstate["wipeStatus"] = 0
            devices[syncstate["deviceid"]] = syncstate
        except Exception as err:
            API.logger.warn("Failed to decode sync state: {}({})".format(type(err).__name__, ", ".join(str(arg) for arg in err.args)))
    if DB.minVersion(93):
        for device in UserDevices.query.filter(UserDevices.userID == userID)\
                                       .with_entities(UserDevices.deviceID, UserDevices.status):
            if device.deviceID in devices:
                devices[device.deviceID]["wipeStatus"] = device.status
            else:
                devices[device.deviceID] = {"deviceid": device.deviceID, "wipeStatus": device.status}
    return jsonify(data=tuple(devices.values()))


def getUsernamesFromFile(domainID, userID, name):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID)\
                      .with_entities(Users.username, Users.maildir).first()
    if user is None:
        return jsonify(message="User not found"), 404
    try:
        with open(user.maildir+"/config/"+name+".txt", encoding="utf-8") as file:
            content = [line.strip() for line in file if line.strip() != ""]
    except (FileNotFoundError, PermissionError, TypeError):
        content = []
    return jsonify(data=content)


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/delegates", methods=["GET"])
@secure(requireDB=True)
def getUserDelegates(domainID, userID):
    return getUsernamesFromFile(domainID, userID, "delegates")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sendas", methods=["GET"])
@secure(requireDB=True)
def getUserSendas(domainID, userID):
    return getUsernamesFromFile(domainID, userID, "sendas")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sendass", methods=["GET"])
def sendAss(*args, **kwargs):
    from random import choices
    from flask import Response
    content = ("                          /\\          /\\\n                         ( \\\\        // )\n                     "
               "     \\ \\\\      // /\n                           \\_\\\\||||//_/\n                            \\/ _  _ \\\n "
               "                          \\/|(O)(O)|\n                          \\/ |      |\n      ___________________\\/  "
               "\\      /\n     //                //     |____|\n    //                ||     /      \\\n   //|               "
               " \\|     \\ 0  0 /\n  // \\       )         V    / \\____/\n //   \\     /        (     /\n     \\   /________"
               "_|  |_/\n       /  /\\   /     |  ||\n      /  / /  /      \\  ||\n      | |  | |        | ||\n      | |  | | "
               "       | ||\n      |_|  |_|        |_||\n       \\_\\  \\_\\        \\_\\\\\n",
               "⠄⠄⠸⣿⣿⢣⢶⣟⣿⣖⣿⣷⣻⣮⡿⣽⣿⣻⣖⣶⣤⣭⡉⠄⠄⠄⠄⠄\n⠄⠄⠄⢹⠣⣛⣣⣭⣭⣭⣁⡛⠻⢽⣿⣿⣿⣿⢻⣿⣿⣿⣽⡧⡄⠄⠄⠄\n⠄⠄⠄⠄⣼⣿⣿⣿⣿⣿⣿⣿⣿⣶⣌⡛⢿⣽⢘⣿⣷⣿⡻⠏⣛⣀⠄⠄\n⠄⠄⠄⣼⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣦⠙⡅"
               "⣿⠚⣡⣴⣿⣿⣿⡆⠄\n⠄⠄⣰⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣷⠄⣱⣾⣿⣿⣿⣿⣿⣿⠄\n⠄⢀⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⢸⣿⣿⣿⣿⣿⣿⣿⣿⠄\n⠄⣸⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡿⠣⣿⣿⣿⣿⣿⣿⣿⣿⣿⠄\n⠄⣿⣿⣿⣿⣿⣿⣿"
               "⣿⣿⣿⣿⣿⣿⠿⠛⠑⣿⣮⣝⣛⠿⠿⣿⣿⣿⣿⠄\n⢠⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣶⠄⠄⠄⠄⣿⣿⣿⣿⣿⣿⣿⣿⣿⡟⠄\n⢸⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠇⠄⠄⠄⠄⢹⣿⣿⣿⣿⣿⣿⣿⣿⠁⠄\n⣸⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠏⠄⠄⠄⠄⠄⠸⣿⣿⣿⣿⣿⡿⢟⣣⣀"
               "\n")
    return Response(choices(content, (10, 1)), headers={"Content-Type": "text/plain"})


def writeUsernamesToFile(domainID, userID, name):
    checkPermissions(DomainAdminPermission(domainID))
    data = request.get_json(silent=True)
    if not isinstance(data, list):
        return jsonify(message="Invalid or missing data"), 400
    for entry in data:
        if not formats.email.match(entry):
            return jsonify(message="Invalid {} e-mail '{}'".format(name, entry))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID)\
                      .with_entities(Users.username, Users.maildir).first()
    if user is None:
        return jsonify(message="User not found"), 404
    try:
        filename = user.maildir+"/config/"+name+".txt"
        with open(filename, "w", encoding="utf-8") as file:
            file.write("\n".join(data)+"\n")
    except (FileNotFoundError, PermissionError) as err:
        return jsonify(message="Failed to write {}: {}".format(name, " - ".join(str(arg) for arg in err.args))), 500
    except TypeError:
        return jsonify(message="User does not support "+name), 400
    try:
        setDirectoryOwner(filename, Config["options"].get("fileUid"), Config["options"].get("fileGid"))
        setDirectoryPermission(filename, Config["options"].get("filePermissions"))
    except Exception:
        pass
    return jsonify(message=name.capitalize()+" updated")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/delegates", methods=["PUT"])
@secure(requireDB=True)
def setUserDelegates(domainID, userID):
    return writeUsernamesToFile(domainID, userID, "delegates")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sendas", methods=["PUT"])
@secure(requireDB=True)
def setUserSendas(domainID, userID):
    return writeUsernamesToFile(domainID, userID, "sendas")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sync/<deviceID>", methods=["DELETE"])
@secure(requireDB=True)
def removeDevice(domainID, userID, deviceID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import DB, UserDevices, Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        client.removeDevice(Config["sync"]["syncStateFolder"], deviceID)
    UserDevices.query.filter(UserDevices.userID == userID, UserDevices.deviceID == deviceID).delete()
    DB.session.commit()
    return jsonify(message="Success")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sync", methods=["DELETE"])
@secure(requireDB=True)
def removeSyncStates(domainID, userID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        client.removeSyncStates(Config["sync"]["syncStateFolder"])
    return jsonify(message="Success")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sync/<deviceID>/resync", methods=["PUT"])
@secure(requireDB=True)
def resyncDevice(domainID, userID, deviceID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        res = client.resyncDevice(Config["sync"]["syncStateFolder"], deviceID, userID)
        return jsonify(message="Great success" if res else "Success")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/sync/<deviceID>/wipe", methods=["POST", "DELETE"])
@secure(requireDB=True, authLevel="user")
def setDeviceWipe(domainID, userID, deviceID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import DB, Users, UserDevices, UserDeviceHistory
    if not DB.minVersion(93):
        return jsonify(message="Database schema too old - please update to at least n93"), 503
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    device = UserDevices.query.filter(UserDevices.userID == userID, UserDevices.deviceID == deviceID).first()
    status = device.status if device is not None else 0
    if (status < 2 and request.method == "DELETE") or \
       (status >= 2 and request.method == "POST"):
        return jsonify(message="Nothing to to")
    if request.method == "DELETE":
        device.status = 1
        DB.session.add(UserDeviceHistory(dict(userDeviceID=device.ID, time=datetime.utcnow(), remoteIP=request.remote_addr,
                                              status=0)))
        DB.session.commit()
        return jsonify(message="Wipe canceled")
    data = request.get_json(silent=True) or {}
    if "password" not in data or not request.auth["user"].chkPw(data["password"]):
        return jsonify(message="User password required"), 403
    if device is None:
        device = UserDevices(dict(userID=userID, deviceID=deviceID, status=2))
        DB.session.add(device)
        DB.session.flush()
    device.status = data.get("status", 2)
    DB.session.add(UserDeviceHistory(dict(userDeviceID=device.ID, time=datetime.utcnow(), remoteIP=request.remote_addr,
                                          status=device.status)))
    DB.session.commit()
    return jsonify(message="Device wipe requested.")


@API.route(api.BaseRoute+"/domains/<int:domainID>/syncPolicy", methods=["GET"])
@secure(requireDB=True, requireAuth="optional")
def getDomainSyncPolicy(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    if domain.syncPolicy is None:
        return jsonify(data=Config["sync"]["defaultPolicy"])
    policy = dict(Config["sync"]["defaultPolicy"])
    policy.update(domain.syncPolicy)
    return jsonify(data=policy)


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/storeAccess", methods=["POST", "PATCH"])
@secure(requireDB=True)
def setUserStoreAccess(domainID, userID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users, DB, UserSecondaryStores
    from sqlalchemy import insert
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    data = request.get_json(silent=True)
    if data is None or "username" not in data:
        return jsonify(message="Invalid data"), 400
    primary = Users.query.with_entities(Users.ID).filter(Users.username == data["username"],
                                                         Users.orgID == user.orgID).first()
    if primary is None:
        return jsonify(message="Could not find user to grant access to"), 404
    eid = makeEidEx(0, PrivateFIDs.IPMSUBTREE)
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        client.setFolderMember(eid, data["username"], Permissions.STOREOWNER)
    if DB.minVersion(91):
        DB.session.execute(insert(UserSecondaryStores).values(primary=primary, secondary=user.ID).prefix_with("IGNORE"))
        DB.session.commit()
    return jsonify(message="Success."), 201 if request.method == "POST" else 200


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/storeAccess", methods=["PUT"])
@secure(requireDB=True)
def setUserStoreAccessMulti(domainID, userID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users, DB, UserSecondaryStores
    from sqlalchemy import insert
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    data = request.get_json(silent=True)
    if data is None or "usernames" not in data:
        return jsonify(message="Invalid data"), 400
    primary = Users.query.filter(Users.username.in_(data["usernames"]), Users.orgID == user.orgID)\
                         .with_entities(Users.ID, Users.username).all()
    eid = makeEidEx(0, PrivateFIDs.IPMSUBTREE)
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        res = client.setFolderMembers(eid, [user.username for user in primary], Permissions.STOREOWNER)
    if DB.minVersion(91):
        UserSecondaryStores.query.filter(UserSecondaryStores.secondaryID == user.ID).delete(synchronize_session=False)
        if len(primary):
            DB.session.execute(insert(UserSecondaryStores).values([{"primary": prim.ID, "secondary": user.ID}
                                                                   for prim in primary]))
        DB.session.commit()
    return jsonify(message="{} user{} updated".format(res, "" if res == 1 else "s"))


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/storeAccess", methods=["GET"])
@secure(requireDB=True)
def getUserStoreAccess(domainID, userID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        memberList = exmdb.FolderMemberList(client.getFolderMemberList(makeEidEx(0, PrivateFIDs.IPMSUBTREE)))
        members = [{"ID": member.id, "displayName": member.name, "username": member.mail} for member in memberList.members
                   if member.rights & Permissions.STOREOWNER]
    return jsonify(data=members)


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/storeAccess/<username>", methods=["DELETE"])
@secure(requireDB=True)
def deleteUserStoreAccess(domainID, userID, username):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import DB, Users, UserSecondaryStores
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no store"), 400
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        client.setFolderMember(makeEidEx(0, PrivateFIDs.IPMSUBTREE), username, Permissions.STOREOWNER, client.REMOVE)
    if DB.minVersion(91):
        primary = Users.query.with_entities(Users.ID).filter(Users.username == username).first()
        if primary is not None:
            UserSecondaryStores.query.filter(UserSecondaryStores.primaryID == primary.ID,
                                             UserSecondaryStores.secondaryID == user.ID).delete()
        DB.session.commit()
    return jsonify(message="Success")


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/oof", methods=["GET"])
@secure(requireDB=True)
def getUserOof(domainID, userID):
    def mkTimeStr(timestamp):
        return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S")

    def getReply(maildir, cat):
        path = os.path.join(maildir, "config", cat)
        try:
            with open(path) as file:
                reply = file.read()
                subject = reply.find("Subject: ")
                subject = None if subject == -1 else reply[subject+9:reply.find("\n", subject)]
                body = reply.find("\n\n")
                return subject, ("" if body == -1 else reply[body+2:])
        except Exception:
            return None, None

    def addIf(name, value):
        if value is not None:
            data[name] = value

    checkPermissions(DomainAdminPermission(domainID))
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).with_entities(Users.maildir).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no maildir"), 400
    data = {"state": 0, "externalAudience": 0}
    path = os.path.join(user.maildir, "config", "autoreply.cfg")
    try:
        parser = configparser.ConfigParser()
        with open(path) as file:
            parser.read_string("[none]\n"+file.read())
    except Exception:
        return jsonify(data)
    oofstate = parser["none"]
    data["state"] = oofstate.getint("oof_state", 0)
    allowExternal = oofstate.getint("allow_external_oof", 0)
    externalAudience = oofstate.getint("external_audience", 0)
    data["externalAudience"] = 0 if not allowExternal else 1 if externalAudience else 2
    if "start_time" in oofstate:
        data["startTime"] = mkTimeStr(oofstate.getint("start_time"))
    if "end_time" in oofstate:
        data["endTime"] = mkTimeStr(oofstate.getint("end_time"))
    subject, reply = getReply(user.maildir, "internal-reply")
    addIf("internalSubject", subject)
    addIf("internalReply", reply)
    subject, reply = getReply(user.maildir, "external-reply")
    addIf("externalSubject", subject)
    addIf("externalReply", reply)
    return jsonify(data)


@API.route(api.BaseRoute+"/domains/<int:domainID>/users/<int:userID>/oof", methods=["PUT"])
@secure(requireDB=True)
def setUserOof(domainID, userID):
    def getTimestamp(timeStr):
        return str(int(time.mktime(datetime.strptime(timeStr, "%Y-%m-%d %H:%M:%S").timetuple())))

    def writeBody(maildir, cat, subject, content):
        path = os.path.join(maildir, "config", cat)
        if content is None:
            return os.unlink(path)
        with open(path, "w", newline="\r\n") as file:
            file.write('Content-Type: text/html;\n\tcharset="utf-8"\n')
            if subject:
                file.write("Subject: "+subject+"\n")
            file.write("\n")
            file.write(content)
        setDirectoryOwner(path, Config["options"].get("fileUid"), Config["options"].get("fileGid"))
        setDirectoryPermission(path, Config["options"].get("filePermissions"))

    checkPermissions(DomainAdminPermission(domainID))
    data = request.get_json(silent=True)
    if not data:
        return jsonify(message="Missing or incomplete data"), 400
    from orm.users import Users
    user = Users.query.filter(Users.ID == userID, Users.domainID == domainID).with_entities(Users.maildir).first()
    if user is None:
        return jsonify(message="User not found"), 404
    if not user.maildir:
        return jsonify(message="User has no maildir"), 400
    config = os.path.join(user.maildir, "config", "autoreply.cfg")
    externalAudience = data.get("externalAudience", 0)
    template = "oof_state = {}\nallow_external_oof = {:d}\nexternal_audience = {:d}\n"
    with open(config, "w") as file:
        file.write(template.format(data.get("state", 0), externalAudience != 0, externalAudience == 1))
        if "startTime" in data:
            file.write("start_time = "+getTimestamp(data["startTime"])+"\n")
        if "endTime" in data:
            file.write("end_time = "+getTimestamp(data["endTime"])+"\n")
    setDirectoryOwner(config, Config["options"].get("fileUid"), Config["options"].get("fileGid"))
    setDirectoryPermission(config, Config["options"].get("filePermissions"))
    if "internalReply" in data:
        writeBody(user.maildir, "internal-reply", data.get("internalSubject"), data["internalReply"])
    if "externalReply" in data:
        writeBody(user.maildir, "external-reply", data.get("externalSubject"), data["externalReply"])
    return jsonify(message="Success")
